/*-
 * Copyright (c) 2005 Valentin Nechayev <netch@netch.kiev.ua>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: struc.c,v 1.7 2005/01/12 21:56:29 netch Exp $
 */

#include <sys/endian.h>
#include <string.h>
#include "struc.h"
#include "assert.h"

/* Reference to crc32.c */
uint32_t crc32(const void *buf, size_t size);

#if _BYTE_ORDER != _LITTLE_ENDIAN

static void
uuid_from_le( uuid_t* local, uuid_t* le )
{
  memmove( local, le, sizeof(uuid_t) );
  local->time_low = le32toh(le->time_low);
  local->time_mid = le16toh(le->time_mid);
  local->time_hi_and_version = le16toh(le->time_hi_and_version);
}

static void
uuid_to_le( uuid_t* local, uuid_t* le )
{
  memmove( local, le, sizeof(uuid_t) );
  le->time_low = htole32(local->time_low);
  le->time_mid = htole16(local->time_mid);
  le->time_hi_and_version = htole16(local->time_hi_and_version);
}

#endif

void
gpt_header_unpack( gpt_header* gh, const unsigned char* buf )
{
  /* Simply copy values */
  assert( sizeof( *gh ) == 92 );
  memmove( (void*)gh, buf, 92 );
#if _BYTE_ORDER != _LITTLE_ENDIAN
  gh->gh_revision = le32toh(*(uint32_t*)&buf[8]);
  gh->gh_hsize = le32toh(*(uint32_t*)&buf[12]);
  gh->gh_reserved = 0;
  gh->gh_mylba = le64toh(*(uint64_t*)&buf[24]);
  gh->gh_altlba = le64toh(*(uint64_t*)&buf[32]);
  gh->gh_fulba = le64toh(*(uint64_t*)&buf[40]);
  gh->gh_lulba = le64toh(*(uint64_t*)&buf[48]);
  uuid_from_le( &gh->gh_guid, (uuid_t*)&buf[56] );
  gh->gh_listlba = le64toh(*(uint64_t*)&buf[72]);
  gh->gh_listcnt = le32toh(*(uint32_t*)&buf[80]);
  gh->gh_listelemsize = le32toh(*(uint32_t*)&buf[84]);
  gh->gh_listcrc = le32toh(*(uint32_t*)&buf[88]);
#endif
}

void
gpt_header_pack( const gpt_header* gh, unsigned char* buf )
{
  /* Simply copy values */
  assert( sizeof( *gh ) == 92 );
  memmove( buf, gh, 92 );
#if _BYTE_ORDER != _LITTLE_ENDIAN
  *(uint32_t*)&buf[8]  = htole32(gh->gh_revision);
  *(uint32_t*)&buf[12] = htole32(gh->gh_hsize);
  *(uint32_t*)&buf[16] = 0;
  *(uint32_t*)&buf[20] = 0;
  *(uint64_t*)&buf[24] = htole64(gh->gh_mylba);
  *(uint64_t*)&buf[32] = htole64(gh->gh_altlba);
  *(uint64_t*)&buf[40] = htole64(gh->gh_fulba);
  *(uint64_t*)&buf[48] = htole64(gh->gh_lulba);
  uuid_to_le( (uuid_t*)&gh->gh_guid, (uuid_t*)&buf[56] );
  *(uint64_t*)&buf[72] = htole64(gh->gh_listlba);
  *(uint32_t*)&buf[80] = htole32(gh->gh_listcnt);
  *(uint32_t*)&buf[84] = htole32(gh->gh_listelemsize);
  *(uint32_t*)&buf[88] = htole32(gh->gh_listcrc);
#endif

  /* Add signature and recalculate checksum */
  memmove( buf, "EFI PART", 8 );
  *(uint32_t*)&buf[12] = 92;
  *(uint32_t*)&buf[16] = 0;
  /* XXX check header padding? */
  *(uint32_t*)&buf[16] = htole32( crc32( buf, 92 ) );
}

int
gpt_header_check_sig( unsigned char* buf )
{
  return !memcmp( buf, "EFI PART", 8 );
}

/* Check whether header CRC is valid. Leave CRC field untouched after
 * calculation.
 */

int
gpt_header_check_crc( unsigned char* buf )
{
  uint32_t* crcpos = (uint32_t*) &buf[16];
  uint32_t* sizepos = (uint32_t*) &buf[12];
  uint32_t crcold = le32toh( *crcpos );
  int ret = ( crcold != crc32( buf, le32toh( *sizepos ) ) );
  *crcpos = htole32( crcold );
  return ret;
}

void
gpt_elem_unpack( gpt_elem* ge, const unsigned char* buf )
{
  memmove( ge, buf, 128 );
#if _BYTE_ORDER != _LITTLE_ENDIAN
  uuid_to_le( &ge->ge_type, buf );
  uuid_to_le( &ge->ge_id, &buf[16] );
  ge->ge_begin = le64toh(*(uint64_t*)&buf[32]);
  ge->ge_end   = le64toh(*(uint64_t*)&buf[40]);
  ge->ge_attr  = le64toh(*(uint64_t*)&buf[48]);
#endif
}

void
gpt_elem_pack( const gpt_elem* ge, unsigned char* buf )
{
  memmove( buf, ge, 128 );
#if _BYTE_ORDER != _LITTLE_ENDIAN
  uuid_to_le( &ge->ge_type, buf );
  uuid_to_le( &ge->ge_id, &buf[16] );
  *(uint64_t*)&buf[32] = htole64(ge->ge_begin);
  *(uint64_t*)&buf[40] = htole64(ge->ge_end);
  *(uint64_t*)&buf[48] = htole64(ge->ge_attr);
#endif
}

int
gpt_array_check_crc( unsigned char* buf, size_t bufsize, uint32_t crc )
{
  return crc32( buf, bufsize ) == crc;
}
